#include "config.h"

const char* ConfigFilePath = "/config.json";
const char* ContactsFilePath = "/contacts.json";

void setup_storage(cfg_t *out) {
    if (!SPIFFS.begin(true)) {
        Serial.println("SPIFFS Mount Failed");
        return;
    }
    if (!SPIFFS.exists(ConfigFilePath)) {
        create_default_config();
    } else {
        read_config(out);
    }
    if (!SPIFFS.exists(ContactsFilePath)) {
        create_default_contactlist();
    }
}
void read_config(cfg_t *out) {
    File file = SPIFFS.open(ConfigFilePath, "r");
    if (!file) {
        Serial.println("Failed to open config file");
        return;
    }

    JsonDocument doc;
    DeserializationError error = deserializeJson(doc, file);
    if (error) {
        Serial.println("Failed to parse config file");
        return;
    }

    String ap_ssid = doc["ap_ssid"].as<String>();
    String ap_pass = doc["ap_pass"].as<String>();
    String wifi1_ssid = doc["wifi1_ssid"].as<String>();
    String wifi1_pass = doc["wifi1_pass"].as<String>();
    String wifi2_ssid = doc["wifi2_ssid"].as<String>();
    String wifi2_pass = doc["wifi2_pass"].as<String>();
    wifi1_pass.toCharArray(out->wifi1_pass, sizeof(out->wifi1_pass));
    wifi1_ssid.toCharArray(out->wifi1_ssid, sizeof(out->wifi1_ssid));
    wifi2_pass.toCharArray(out->wifi2_pass, sizeof(out->wifi2_pass));
    wifi2_ssid.toCharArray(out->wifi2_ssid, sizeof(out->wifi2_ssid));
    ap_ssid.toCharArray(out->ap_ssid, sizeof(out->ap_ssid));
    ap_pass.toCharArray(out->ap_pass, sizeof(out->ap_pass));
    //
    //
    //
    float freq = doc["tx_freq"].as<float>();
    float baud = doc["tx_baud"].as<float>();
    float dev = doc["tx_dev"].as<float>();
    int power = doc["tx_power"].as<int>();
    bool empty_queue = doc["tx_empty_queue"].as<bool>();
    out->tx_freq = freq;
    out->tx_baud = baud;
    out->tx_dev = dev;
    out->tx_power = power;
    out->tx_empty_queue = empty_queue;
    //
    //
    //
    int mqtt_port = doc["mqtt_port"].as<int>();
    String mqtt_host = doc["mqtt_host"].as<String>();
    String mqtt_user = doc["mqtt_user"].as<String>();
    String mqtt_pass = doc["mqtt_pass"].as<String>();
    String mqtt_topic = doc["mqtt_topic"].as<String>();
    mqtt_host.toCharArray(out->mqtt_host, sizeof(out->mqtt_host));
    mqtt_user.toCharArray(out->mqtt_user, sizeof(out->mqtt_user));
    mqtt_pass.toCharArray(out->mqtt_pass, sizeof(out->mqtt_pass));
    mqtt_topic.toCharArray(out->mqtt_topic, sizeof(out->mqtt_topic));
    out->mqtt_port = mqtt_port;
    //
    //
    //
    int broadcast_ric = doc["broadcast_ric"].as<int>();
    int broadcast_fun = doc["broadcast_fun"].as<int>();
    out->broadcast_ric = broadcast_ric;
    out->broadcast_fun = broadcast_fun;
    //
    int dwd_interval = doc["dwd_interval"].as<int>();
    bool dwd_enable = doc["dwd_enable"].as<bool>();
    String dwd_region = doc["dwd_region"].as<String>();
    out->dwd_interval = dwd_interval;
    out->dwd_enable = dwd_enable;
    dwd_region.toCharArray(out->dwd_region, sizeof(out->dwd_region));
    if (doc["dwd_fun"].is<int>()) {
        int dwd_fun = doc["dwd_fun"].as<int>();
        out->dwd_fun = dwd_fun;
    }
    //
    int mowas_interval = doc["mowas_interval"].as<int>();
    bool mowas_enable = doc["mowas_enable"].as<bool>();
    String mowas_region = doc["mowas_region"].as<String>();
    out->mowas_interval = mowas_interval;
    out->mowas_enable = mowas_enable;
    mowas_region.toCharArray(out->mowas_region, sizeof(out->mowas_region));
    if (doc["mowas_fun"].is<int>()) {
        int mowas_fun = doc["mowas_fun"].as<int>();
        out->mowas_fun = mowas_fun;
    }
    //
    //
    if (doc["time_interval"].is<int>()) {
        int time_interval = doc["time_interval"].as<int>();
        bool time_enable = doc["time_enable"].as<bool>();
        out->time_enable = time_enable;
        out->time_interval = time_interval;
        int time_ric = doc["time_ric"].as<int>();
        int time_fun = doc["time_fun"].as<int>();
        out->time_ric = time_ric;
        out->time_fun = time_fun;

        int time_mode = doc["time_mode"].as<int>();
        out->time_mode = time_mode;

        if (doc["time_zone"].is<String>()) {
            String time_zone = doc["time_zone"].as<String>();
            time_zone.toCharArray(out->time_zone, sizeof(out->time_zone));
        }
    }
    if (doc["idle_interval"].is<int>()) {
        int idle_interval = doc["idle_interval"].as<int>();
        bool idle_enable = doc["idle_enable"].as<bool>();
        int idle_mode = doc["idle_mode"].as<int>();
        out->idle_enable = idle_enable;
        out->idle_interval = idle_interval;
        out->idle_mode = idle_mode;
    }
    if (doc["pocsag_german"].is<bool>()) {
        bool pocsag_german = doc["pocsag_german"].as<bool>();
        out->pocsag_german = pocsag_german;
    }
    if (doc["device_name"].is<String>()) {
        String device_name = doc["device_name"].as<String>();
        device_name.toCharArray(out->device_name, sizeof(out->device_name));
    }
    if (doc["oled_timeout"].is<int>()) {
        int oled_timeout = doc["oled_timeout"].as<int>();
        out->oled_timeout = oled_timeout;
    }
    Serial.println("Loaded Configuration:");
    Serial.printf("AP SSID: %s\n", out->ap_ssid);
    Serial.printf("AP Password: %s\n", out->ap_pass);
    Serial.printf("1. SSID: '%s'\n", out->wifi1_ssid);
    Serial.printf("1. Password: '%s'\n", out->wifi1_pass);
    Serial.printf("2. SSID: '%s'\n",out->wifi2_ssid);
    Serial.printf("2. Password: '%s'\n",out->wifi2_pass);
    Serial.printf("MQTT Host/Port: '%s':%d\n", out->mqtt_host, out->mqtt_port);
    Serial.printf("User/Password: '%s':'%s'\n",out->mqtt_user, out->mqtt_pass);
    Serial.printf("Topic: '%s'\n",out->mqtt_topic);
    Serial.println();
    Serial.printf("German POCSAG: %d\n", out->pocsag_german);
    Serial.printf("Transmitter: %f MHz @%dbd - dev=%dkHz\n", out->tx_freq, out->tx_baud, out->tx_dev);
    Serial.printf("Broadcast: RIC %d fun=%d\n", out->broadcast_ric, out->broadcast_fun);
    Serial.printf("DWD: Enable=%d Interval=%dm SubRIC=%d Region=%s\n", out->dwd_enable, out->dwd_interval, out->dwd_fun, out->dwd_region);
    Serial.printf("MoWaS: Enable=%d Interval=%dm SubRIC=%d Region=%s\n", out->mowas_enable, out->mowas_interval, out->mowas_fun, out->mowas_region);
    Serial.printf("Time: Enable=%d Interval=%dm RIC=%d SubRIC=%d Mode=%d Zone=%s\n", out->time_enable, out->time_interval, out->time_ric, out->time_fun, out->time_mode, out->time_zone);
    //
    doc.clear();
    file.close();
}
String cfg_tostring() {
    File file = SPIFFS.open(ConfigFilePath, "r");
    if (!file) {
        Serial.println("Failed to open config file");
        return "";
    }
    String output = file.readString();
    file.close();
    return output;
}
void create_default_config() {
    Serial.println("creating default config");

    JsonDocument doc;

    doc["wifi1_ssid"] = "";
    doc["wifi1_pass"] = "";
    doc["wifi2_ssid"] = "";
    doc["wifi2_pass"] = "";
    doc["ap_ssid"] = "POCSAG";
    doc["ap_pass"] = "Motorola";
    // modem
    doc["tx_freq"] = 434.230f;
    doc["tx_baud"] = 1200;
    doc["tx_dev"] = 4.5;
    doc["tx_power"] = 13;
    doc["tx_empty_queue"] = false;
    // mqtt
    doc["mqtt_host"] = "";
    doc["mqtt_user"] = "";
    doc["mqtt_pass"] = "";
    doc["mqtt_topic"] = "";
    doc["mqtt_port"] = 1883;
    //
    doc["broadcast_ric"] = 9999;
    doc["broadcast_fun"] = 3;
    //
    doc["dwd_fun"] = 3;
    doc["dwd_interval"] = 5;
    doc["dwd_enable"] = false;
    doc["dwd_region"] = "";
    doc["mowas_fun"] = 3;
    doc["mowas_interval"] = 5;
    doc["mowas_enable"] = false;
    doc["mowas_region"] = "";
    
    doc["time_enable"] = false;
    doc["time_interval"] = 720;
    doc["time_ric"] = 0;
    doc["time_fun"] = 3;
    doc["time_mode"] = 0;
    doc["time_zone"] = "Europe/Berlin";

    doc["idle_enable"] = false;
    doc["idle_interval"] = 1;
    doc["idle_mode"] = 0;

    doc["pocsag_german"] = false;
    doc["oled_timeout"] = 2;
    doc["device_name"] = "smartPOC";
    File file = SPIFFS.open(ConfigFilePath, "w");
    if (!file) {
        Serial.println("Failed to open config file for writing");
        return;
    }

    serializeJson(doc, file);
    file.close();
    doc.clear();

    Serial.println("Configuration saved.");
}
//
void create_default_contactlist() {
    Serial.println("creating default contacts");

    JsonArray jArr;
    JsonObject testContact;
    testContact["r"] = 133701;
    testContact["n"] = "Contact Name";
    testContact["f"] = 3;
    testContact["tl"] = 80;
    testContact["t"] = "Test";
    jArr.add(testContact);
    //

    File file = SPIFFS.open(ContactsFilePath, "w");
    if (!file) {
        Serial.println("Failed to open config file for writing");
        return;
    }

    serializeJson(jArr, file);
    file.close();
    jArr.clear();

    Serial.println("Contacts saved.");
}
String contacts_tostring() {
    File file = SPIFFS.open(ContactsFilePath, "r");
    if (!file) {
        Serial.println("Failed to open contacts file");
        return "";
    }
    String output = file.readString();
    file.close();
    return output;
}
void contacts_write(String data) {
    File file = SPIFFS.open(ContactsFilePath, "w");
    if (!file) {
        Serial.println("Failed to open contacts file for writing");
        return;
    }
    file.print(data);
    file.close();
    Serial.println("Contacts saved.");
}
//
JsonDocument transactionJson;
void cfg_startTransaction() {
    File file = SPIFFS.open(ConfigFilePath, "r");
    if (!file) {
        Serial.println("Failed to open config file");
        return;
    }
    DeserializationError error = deserializeJson(transactionJson, file);
    if (error) {
        Serial.println("Failed to parse config file");
        return;
    }
    file.close();
}
void cfg_adjust(String key, bool val) {
    Serial.printf("adjustng cfg %s\n", key);
    transactionJson[key] = val;
}
void cfg_adjust(String key, String val) {
    Serial.printf("adjustng cfg %s = '%s'\n", key, val);
    transactionJson[key] = val;
}
void cfg_adjust(String key, float val) {
    Serial.printf("adjustng cfg %s = %f\n", key, val);
    transactionJson[key] = val;
}
void cfg_adjust(String key, int val) {
    Serial.printf("adjustng cfg %s = %d\n", key, val);
    transactionJson[key] = val;
}
void cfg_write() {
    File file = SPIFFS.open(ConfigFilePath, "w");
    if (!file) {
        Serial.println("Failed to open config file for writing");
        return;
    }
    serializeJson(transactionJson, file);
    file.close();
    //transactionJson.clear();
    Serial.println("wrote cfg");
}